from __future__ import annotations

import collections as C
import typing as T

from lazyfpl import structures


def team_constraint(lineup: T.Sequence[structures.Player], n: int) -> bool:
    """
    Checks if any team is over-represented in the lineup.

    Returns True if the maximum number of players from any
    single team in the lineup does not exceed 'n'.
    """
    return max(C.Counter(p.team for p in lineup).values()) <= n


def gkp_def_same_team(
    lineup: T.Sequence[structures.Player],
) -> bool:
    """
    Determines if any goalkeepers (GKP) and defenders (DEF) in the
    lineup are from the same team.
    Returns True if there is at least one common team between
    goalkeepers and defenders.
    """
    gkps = {p.team for p in lineup if p.position == "GKP"}
    defs = {p.team for p in lineup if p.position == "DEF"}
    return bool(gkps.intersection(defs))



from __future__ import annotations

import dataclasses
import datetime
import statistics
import traceback
import typing

import more_itertools
import torch

from lazyfpl import conf, fetch, helpers, ml_model, structures


@dataclasses.dataclass(frozen=True)
class PredictionOutcome:
    """
    Represents the outcome of a prediction, including predicted and
    actual values, and kickoff time.
    """

    prediceted: float
    target: float
    kickoff: datetime.datetime


def backeval(
    player: structures.Player,
    backtrace: int = conf.backtrace,
    backstep: int = 10,
) -> typing.Iterator[PredictionOutcome]:
    """
    Back-evaluates a player's performance predictions for a
    given number of past fixtures.
    """
    with torch.no_grad():
        net = ml_model.load_model(player).eval()
        fixutres = [f for f in player.fixutres if not f.upcoming][-backstep:]
        for *context, target in more_itertools.sliding_window(fixutres, backtrace + 1):
            xP = (
                net(
                    torch.tensor(
                        [(ml_model.features(c)).flattend() for c in context],
                        dtype=torch.float32,
                    ).unsqueeze(0)
                )
                .detach()
                .numpy()
            )

            assert xP is not None
            assert target.points is not None
            yield PredictionOutcome(
                prediceted=xP,
                target=target.points,
                kickoff=target.kickoff_time,
            )


def players_backeval(
    top_n_total_points: int = 100,
) -> typing.Generator[
    tuple[structures.Player, tuple[PredictionOutcome, ...]],
    None,
    None,
]:
    """
    Evaluates the back-evaluation for all players and returns
    a dictionary with the results.
    """
    for player in sorted(
        sorted(fetch.players(), key=lambda x: x.tp())[-top_n_total_points:],
        key=lambda x: (
            x.team,
            helpers.position_order(x.position),
            x.name,
        ),
    ):
        try:
            yield player, tuple(backeval(player))
        except ValueError as e:
            if conf.debug:
                traceback.print_exception(e)


if __name__ == "__main__":
    player_xp = tuple(players_backeval())

    for player, ev in player_xp:
        print(f"\n{player.name}({player.webname}) - {player.position} - {player.team}")
        for e in sorted(ev, key=lambda x: x.kickoff):
            print(
                f"  When: {e.kickoff.date()} xP: {e.prediceted:<6.1f} "
                f"TP: {e.target:<6.1f} Err: {abs(e.prediceted - e.target):<6.1f}"
            )

    print()
    rms = (
        statistics.mean(
            (v.prediceted - v.target) ** 2 for _, values in player_xp for v in values
        )
        ** 0.5
    )
    print(f"RMS: {rms:.1f}")
    for n in range(1, 6):
        rc = statistics.mean(
            1 if abs(v.prediceted - v.target) <= n else 0
            for _, values in player_xp
            for v in values
        )
        print(f"RC{n}: {rc*100:.1f}")

    print()

    print(f"Num selected for backeval: {len(player_xp)}")



def help_message() -> None:
    print(
        """
    LazyFPL - Fantasy Premier League Team Optimizer - Win at FPL with Laziness

    - Constraints Module
        Contains functions to check various constraints like team composition
        and budget limits.

    - Database Module
        Manages interactions with the database, including queries and data storage.

    - Fetch Module
        Retrieves various types of data related to players, teams, and games.

    - Helpers Module
        Provides utility functions for calculations and data manipulations.

    - Populator Module
        Responsible for populating the database with current and historical FPL data.

    - Structures Module
        Defines data structures and models for representing players, teams, and games.

    - Transfer Modulee
        Assists in making transfer decisions based on various strategies
        and constraints.

    - Differentials Module
        Offers analytics to identify potential differential picks in
        your fantasy football team.

    Usage:
        Each module is designed to be used as a standalone tool or in conjunction
    with others to enhance your Fantasy Premier League strategy. Use the respective
    module's functions to fetch data, analyze team compositions, make
    transfers, and more.

    For more detailed information on each module, refer to their individual
    documentation or use help(module_name).
    """
    )


if __name__ == "__main__":
    help_message()



from __future__ import annotations

import argparse
import concurrent.futures
import csv
import datetime
import functools
import io
import urllib.parse as up

import pytz
import requests
from dateutil.parser import parse as dtparser
from tqdm.std import tqdm

from lazyfpl import database, structures


def now_tz_utc() -> datetime.datetime:
    """Returns the current time in UTC timezone."""
    return datetime.datetime.now(tz=pytz.utc)


@functools.cache
def bootstrap() -> dict:
    """Fetches data from the Fantasy Premier League API's bootstrap-static endpoint."""
    return requests.get(
        "https://fantasy.premierleague.com/api/bootstrap-static/"
    ).json()


@functools.cache
def summary(id: int) -> dict:
    """Fetches player summary for a given player ID from the
    Fantasy Premier League API."""
    assert isinstance(id, int)
    return requests.get(
        f"https://fantasy.premierleague.com/api/element-summary/{id}/"
    ).json()


@functools.cache
def db_name_pid() -> dict[str, int]:
    """Creates a dictionary mapping player names to their database IDs."""
    return {
        row["name"]: row["id"]
        for row in database.execute(
            """
        SELECT
            id,
            name
        FROM
            player
        """
        )
    }


@functools.cache
def player_id_fuzzer(name: str) -> int | None:
    """Attempts to find a player ID in the database matching the
    given name, with some flexibility for variations."""
    for dname, pid in db_name_pid().items():
        if dname.casefold() == name.casefold():
            return pid
        if all(n in dname.casefold() for n in name.casefold().split()):
            return pid
        if all(n in name.casefold() for n in dname.casefold().split()):
            return pid

    return None


def session_from_url(url: str) -> str:
    """Extracts the session year from a given URL."""
    return up.urlparse(url).path.split("/")[5]


@functools.cache
def past_team_lists() -> dict[str, list[structures.Team]]:
    """Fetches and parses past team data from external sources for different seasons."""
    urls = (
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2023-24/teams.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2022-23/teams.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2021-22/teams.csv",
    )

    return {
        session_from_url(url): [
            structures.Team.model_validate(t)
            for t in csv.DictReader(io.StringIO(requests.get(url).text))
        ]
        for url in urls
    }


@functools.cache
def past_team_lookup(tid: int, session: database.SESSIONS) -> str:
    """Finds the team name for a given team ID and session."""
    assert isinstance(tid, int)
    for team in past_team_lists()[session]:
        if team.id == tid:
            return team.name
    raise ValueError(f"No match: {tid} / {session}.")


@functools.cache
def past_game_lists() -> dict[str, list[dict]]:
    """Fetches and parses past game data from external sources for different seasons."""
    urls = (
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2023-24/gws/merged_gw.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2022-23/gws/merged_gw.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2021-22/gws/merged_gw.csv",
    )
    return {
        session_from_url(url): list(csv.DictReader(io.StringIO(requests.get(url).text)))
        for url in urls
    }


def initialize_database() -> None:
    """Initializes the database with the necessary tables."""
    database.execute(
        """
        CREATE TABLE team (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            short_name TEXT NOT NULL,
            web_team_id INTEGER NOT NULL,
            session TEXT NOT NULL,
            strength INTEGER NOT NULL,
            UNIQUE(name, session)
        )
    """
    )
    database.execute(
        """
        CREATE TABLE player (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            webname TEXT NOT NULL,
            name TEXT NOT NULL,
            price INTEGER NOT NULL,
            team_id INTEGER NOT NULL,
            news TEXT NOT NULL,
            model BLOB,
            FOREIGN KEY(team_id) REFERENCES team(id),
            UNIQUE(webname, name, team_id)
        );
    """
    )
    database.execute(
        """
        CREATE TABLE game (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            is_home INTEGER NOT NULL,
            kickoff REAL NOT NULL,
            minutes INTEGER,
            opponent INTEGER NOT NULL,
            player_id INTEGER NOT NULL,
            points INTEGER,
            position TEXT NOT NULL,
            session TEXT NOT NULL,
            team INTEGER NOT NULL,
            upcoming INTEGER NOT NULL,
            selected INTEGER NOT NULL,
            FOREIGN KEY(opponent) REFERENCES team(id),
            FOREIGN KEY(player_id) REFERENCES player(id),
            FOREIGN KEY(team) REFERENCES team(id)
        );
    """
    )


def nuke_database() -> None:
    """Removes all existing tables from the database."""
    database.execute("""DROP TABLE IF EXISTS game;""")
    database.execute("""DROP TABLE IF EXISTS player;""")
    database.execute("""DROP TABLE IF EXISTS team;""")


def populate_teams() -> None:
    """Populates the database with team data."""
    sql = """
        INSERT INTO team (
            name,
            short_name,
            session,
            web_team_id,
            strength
        ) VALUES (?, ?, ?, ?, ?);
    """
    for session, teams in tqdm(
        past_team_lists().items(),
        ascii=True,
        desc="Populates teams        ",
        ncols=80,
        unit_scale=True,
    ):
        for team in teams:
            database.execute(
                sql,
                (
                    team.name,
                    team.short_name,
                    session,
                    team.id,
                    team.strength,
                ),
            )


def populate_players(session: database.SESSIONS = database.CURRENT_SESSION) -> None:
    """Populates the database with player data for the specified session."""
    sql = """
        INSERT INTO player(
            webname,
            name,
            price,
            news,
            team_id
        ) VALUES (
            ?, ?, ?, ?,
            (SELECT id FROM team WHERE session = ? AND web_team_id = ?)
        );
    """
    for ele in tqdm(
        bootstrap()["elements"],
        ascii=True,
        desc="Populate players       ",
        ncols=80,
        unit_scale=True,
    ):
        database.execute(
            sql,
            (
                ele["web_name"],
                f'{ele["first_name"]} {ele["second_name"]}',
                ele["now_cost"],
                ele["news"],
                session,
                ele["team"],
            ),
        )


def populate_games() -> None:
    """Populates the database with game data, including historic and upcoming games."""
    game_sql = """
        INSERT INTO game (
            session,
            upcoming,
            is_home,

            kickoff,
            minutes,
            points,

            position,
            player_id,
            team,
            opponent,

            selected
        ) VALUES (
            ?, ?, ?, ?, ?, ?, ?,
            ?,
            (SELECT id FROM team WHERE session = ? AND name = ?),
            (SELECT id FROM team WHERE session = ? AND name = ?),
            ?
        );
    """
    for session, games in past_game_lists().items():
        to_insert = []
        for fixture in tqdm(
            games,
            ascii=True,
            desc=f"Populate games: {session}",
            ncols=80,
            unit_scale=True,
        ):
            historic = structures.HistoricGame.model_validate(fixture)
            if pid := player_id_fuzzer(historic.name):
                to_insert.append(
                    (
                        session,
                        False,
                        historic.was_home,
                        historic.kickoff_time,
                        historic.minutes,
                        historic.total_points,
                        historic.position,
                        pid,
                        session,
                        historic.team,
                        session,
                        past_team_lookup(historic.opponent_team, session),
                        historic.selected,
                    ),
                )
        database.executemany(game_sql, tuple(to_insert))

    now = now_tz_utc()
    upcoming_games: list[int] = [
        e["id"] for e in bootstrap()["events"] if dtparser(e["deadline_time"]) > now
    ]

    jobs = list[concurrent.futures.Future]()

    def pooled_summary(
        id: int,
        fullname: str,
        team: str,
    ) -> tuple[dict, str, str]:
        return summary(id)["fixtures"], fullname, team

    with concurrent.futures.ThreadPoolExecutor() as pool:
        for ele in bootstrap()["elements"]:
            jobs.append(
                pool.submit(
                    pooled_summary,
                    ele["id"],
                    " ".join((ele["first_name"], ele["second_name"])),
                    upcoming_team_id_to_name(ele["team"]),
                )
            )

        for done in tqdm(
            concurrent.futures.as_completed(jobs),
            ascii=True,
            desc="Populate upcoming games",
            ncols=80,
            total=len(jobs),
            unit_scale=True,
        ):
            fixtures, fullname, team = done.result()
            for fixture in fixtures:
                upcoming = structures.UpcommingGame.model_validate(fixture)
                if not upcoming.event_name:
                    continue

                if upcoming.event in upcoming_games:
                    team_h = upcoming_team_id_to_name(upcoming.team_h)
                    team_a = upcoming_team_id_to_name(upcoming.team_a)
                    is_home = team == team_h
                    opponent = list({team, team_a, team_h} - {team})[0]
                    database.execute(
                        game_sql,
                        (
                            database.CURRENT_SESSION,
                            True,
                            is_home,
                            upcoming.kickoff_time,
                            None,
                            None,
                            upcoming_position(fullname),
                            player_id_fuzzer(fullname),
                            database.CURRENT_SESSION,
                            team,
                            database.CURRENT_SESSION,
                            opponent,
                            -1,
                        ),
                    )


@functools.cache
def upcoming_team_id_to_name(id: int) -> str:
    """Converts a team's ID to its name for upcoming games."""
    for item in bootstrap()["teams"]:
        if id == item["id"]:
            return item["name"]
    raise ValueError(f"No team name: {id}")


@functools.cache
def upcoming_position(name: str) -> database.POSITIONS:
    """Determines the position of a player based on their name for upcoming games."""
    for element in bootstrap()["elements"]:
        if f'{element["first_name"]} {element["second_name"]}' == name:
            element_type = bootstrap()["element_types"][element["element_type"] - 1]
            return element_type["singular_name_short"]
    raise ValueError(f"No player named: {name}")


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Populator",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.parse_args()

    nuke_database()
    initialize_database()
    populate_teams()
    populate_players()
    populate_games()


if __name__ == "__main__":
    main()



from __future__ import annotations

import dataclasses
import datetime
import statistics
from typing import Generator, Sequence

import pydantic

from lazyfpl import conf, database, helpers


@dataclasses.dataclass(frozen=True)
class Fixture:
    at_home: bool
    kickoff_time: datetime.datetime
    minutes: int | None
    opponent: str
    opponent_short: str
    opponent_strength: int
    player: str
    points: int | None
    session: database.SESSIONS
    team: str
    team_short: str
    team_strength: int
    upcoming: bool
    webname: str


@dataclasses.dataclass(eq=True, unsafe_hash=True)
class Player:
    fixutres: list[Fixture] = dataclasses.field(compare=False, repr=False)
    name: str = dataclasses.field(compare=True)
    news: str = dataclasses.field(compare=True)
    position: database.POSITIONS = dataclasses.field(compare=False)
    price: int = dataclasses.field(compare=False)
    selected: int | None = dataclasses.field(compare=False)
    team: str = dataclasses.field(compare=False)
    team_short: str = dataclasses.field(compare=False)
    webname: str = dataclasses.field(compare=True)
    xP: float | None

    def tp(self, session: database.SESSIONS = database.CURRENT_SESSION) -> int:
        return sum(f.points or 0 for f in self.fixutres if f.session == session)

    def mtm(self, last: int = conf.backtrace) -> float:
        # fixutre.minutes is None if the gameweek is not started
        # need to filter out fixutres with minutes as none.
        try:
            return statistics.mean(
                (
                    f.minutes or 0
                    for f in sorted(
                        (f for f in self.fixutres if not f.upcoming),
                        key=lambda x: x.kickoff_time,
                    )[-last:]
                )
            )
        except statistics.StatisticsError:
            return 0.0

    def upcoming_difficulty(self) -> int:
        upcoming = [f for f in self.fixutres if f.upcoming][: conf.lookahead]
        return sum(f.team_strength - f.opponent_strength for f in upcoming)

    def upcoming_opponents(self) -> list[str]:
        return [
            x.opponent_short
            for x in sorted(
                (f for f in self.fixutres if f.upcoming),
                key=lambda f: f.kickoff_time,
            )
        ]

    def str_upcoming_opponents(self) -> str:
        return " - ".join(self.upcoming_opponents()[: conf.lookahead])

    def __str__(self) -> str:
        raise NotImplementedError

    def display(self) -> dict[str, float | str | None]:
        return {
            "xP": self.xP,
            "Price": self.price,
            "TP": self.tp(),
            "UD": self.upcoming_difficulty(),
            "Selected": self.selected,
            "Webname": self.webname,
            "Team": self.team,
            "Position": self.position,
            "Upcoming opponents": self.str_upcoming_opponents(),
            "News": self.news,
        }


class Squad:
    def __init__(self, players: Sequence[Player]):
        self.players = players

    def price(self) -> int:
        return helpers.squad_price(self.players)

    def best_lineup(self) -> list[Player]:
        return helpers.best_lineup(self.players)

    def SxP(self) -> float:
        # Squad xP
        return helpers.squad_xP(self.players)

    def LxP(self) -> float:
        # Lineup xP
        return helpers.squad_xP(self.best_lineup())

    def CxP(self) -> float:
        # Combined xP
        return (self.SxP() ** 2 + self.LxP() ** 2) ** 0.5

    def tscore(self) -> int:
        return helpers.tcnt(self.players)

    def sscore(self, n: int = conf.lookahead) -> int:
        # "sscore -> "schedule score"
        return helpers.sscore(self.players, n=n)

    def tsscore(self, n: int = conf.lookahead) -> float:
        return helpers.tsscore(self.players)

    def __iter__(self) -> Generator[Player, None, None]:
        yield from self.players

    def __len__(self) -> int:
        return len(self.players)

    def __str(self) -> Generator[str, None, None]:
        yield f"Price: {self.price()/10} Size: {len(self.players)}"
        yield f"LxP: {self.LxP():.1f} SxP: {self.SxP():.1f} CxP: {self.CxP():.1f}"
        yield (
            f"Schedule score: {self.sscore()} Team score: {self.tscore()} "
            f"TSscore: {self.tsscore():.2f}"
        )
        bis = helpers.best_lineup(self.players)
        yield helpers.tabulater(
            [
                dict({"BIS": "X" if p in bis else ""} | p.display())
                for p in sorted(
                    self.players,
                    key=lambda x: (-helpers.position_order(x.position), -(x.xP or 0)),
                )
            ],
        )

    def __str__(self) -> str:
        return "\n".join(self.__str())


class HistoricGame(pydantic.BaseModel):
    assists: int
    bonus: int
    bps: int
    clean_sheets: int
    creativity: float
    element: int
    fixture: float
    goals_conceded: float
    goals_scored: float
    GW: int
    ict_index: float
    influence: float
    kickoff_time: datetime.datetime
    minutes: int
    name: str
    opponent_team: int
    own_goals: int
    penalties_missed: int
    penalties_saved: int
    position: database.POSITIONS
    red_cards: int
    round: int
    saves: int
    selected: int
    team_a_score: int
    team_h_score: int
    team: str
    threat: float
    total_points: int
    transfers_balance: int
    transfers_in: int
    transfers_out: int
    value: int
    was_home: bool
    xP: float
    yellow_cards: int

    @pydantic.field_validator("position", mode="before")
    def gk_to_gkp(cls, value: str) -> str:
        return "GKP" if value == "GK" else value


class UpcommingGame(pydantic.BaseModel):
    code: int
    difficulty: int
    event_name: str | None = pydantic.Field(default=None)
    event: int | None = pydantic.Field(default=None)
    finished: bool
    id: int
    is_home: bool
    kickoff_time: datetime.datetime | None = pydantic.Field(default=None)
    minutes: int
    provisional_start_time: bool
    team_a_score: int | None = pydantic.Field(default=None)
    team_a: int
    team_h_score: int | None = pydantic.Field(default=None)
    team_h: int


class Team(pydantic.BaseModel):
    code: int
    draw: int
    id: int
    loss: int
    name: str
    played: int
    points: int
    position: int
    pulse_id: int
    short_name: str
    strength: int
    unavailable: bool
    win: int



from __future__ import annotations

import argparse
import itertools

from lazyfpl import fetch, helpers

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--top",
        "-t",
        default=None,
        help="Top N players per position. (default: %(default)s)",
        type=int,
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="Drop players with news attched to them. (default: %(default)s)",
    )
    args = parser.parse_args()

    players = sorted(
        fetch.players(),
        key=lambda x: (helpers.position_order(x.position), x.xP or 0),
        reverse=True,
    )

    if args.no_news:
        players = [p for p in players if not p.news]

    players = list(
        itertools.chain.from_iterable(
            [
                list(p)[: args.top]
                for _, p in itertools.groupby(
                    players,
                    key=lambda x: helpers.position_order(x.position),
                )
            ]
        )
    )
    bis = helpers.best_lineup(players)

    print(
        helpers.tabulater(
            [dict({"BIS": "X" if p in bis else ""} | p.display()) for p in players],
        ),
    )



from __future__ import annotations

import dataclasses
import datetime
import functools
import itertools
import traceback

import dateutil.parser
import requests

from lazyfpl import conf, database, ml_model, structures


@dataclasses.dataclass(frozen=True)
class Persona:
    """
    Represents a Persona with first name, second name, and webname.
    """

    first: str
    second: str
    webname: str

    @property
    def combined(self) -> str:
        """
        Returns a combined string of the first and second name.
        """
        return " ".join((self.first, self.second))


@functools.cache
def bootstrap() -> dict:
    """
    Fetches and returns data from the Fantasy Premier League API's
    bootstrap-static endpoint.
    """
    return requests.get(
        "https://fantasy.premierleague.com/api/bootstrap-static/"
    ).json()


def bootstrap_events() -> list[dict]:
    """
    Extracts and returns event-related data from the bootstrap data.
    """
    return bootstrap()["events"]


@functools.cache
def next_gw() -> int:
    """
    Determines and returns the next game week's ID.
    """
    for e in bootstrap_events():
        if e["is_next"]:
            return int(e["id"])
    raise ValueError


@functools.cache
def current_gw() -> int:
    """
    Determines and returns the current game week's ID.
    """
    for e in bootstrap_events():
        if e["is_current"]:
            return int(e["id"])
    raise ValueError


@functools.cache
def next_deadline() -> datetime.timedelta:
    for e in bootstrap_events():
        if e["is_next"]:
            deadline = dateutil.parser.parse(e["deadline_time"])
            now = datetime.datetime.now(tz=datetime.timezone.utc)
            return deadline - now
    raise ValueError


@functools.cache
def person(pid: int) -> Persona:
    """
    Retrieves and returns a Persona object for a given player ID.
    """
    for element in bootstrap()["elements"]:
        if element["id"] == pid:
            return Persona(
                first=element["first_name"],
                second=element["second_name"],
                webname=element["web_name"],
            )
    raise ValueError(f"No player: {pid}")


@functools.cache
def players() -> list[structures.Player]:
    """
    Creates and returns a list of Player objects, each representing a football player.
    """
    pool = list[structures.Player]()

    for (name, webname), _games in itertools.groupby(
        sorted(
            database.games(),
            key=lambda x: (x.player, x.webname),
        ),
        lambda x: (x.player, x.webname),
    ):
        games = sorted(_games, key=lambda x: x.kickoff)
        fixtures = [
            structures.Fixture(
                at_home=game.is_home,
                kickoff_time=game.kickoff,
                minutes=game.minutes,
                opponent=game.opponent,
                opponent_short=game.opponent_short,
                player=name,
                points=game.points,
                session=game.session,
                team=game.team,
                team_short=game.team_short,
                upcoming=game.upcoming,
                webname=webname,
                team_strength=game.team_strength,
                opponent_strength=game.opponent_strength,
            )
            for game in games
        ]

        try:
            next_upcoming = [g for g in games if g.upcoming][0]
            last_game = [g for g in games if not g.upcoming][-1]
        except IndexError as e:
            if conf.debug:
                traceback.print_exception(e)
            continue
        pool.append(
            structures.Player(
                fixutres=fixtures,
                name=name,
                news=games[-1].news,
                position=games[-1].position,
                price=database.price(games[-1].player_id),
                team=next_upcoming.team,
                team_short=next_upcoming.team_short,
                webname=database.webname(games[-1].player_id),
                xP=None,
                selected=last_game.selected,
            )
        )

    for p in pool:
        try:
            p.xP = ml_model.xP(p)
        except ValueError as e:
            if conf.debug:
                traceback.print_exception(e)

    return pool


@functools.cache
def picks() -> list[Persona]:
    """
    Fetches and returns the current team picks from the Fantasy Premier League API.
    """
    if not conf.teamid or not conf.profile:
        raise RuntimeError(
            "Env. FPL-teamid and FPL-cookie/profile must be set. FPL-team id "
            "from URL and cookie 'pl_profile' from 'application' in chrome."
        )

    response = requests.get(
        f"https://fantasy.premierleague.com/api/my-team/{conf.teamid}/",
        cookies={"pl_profile": conf.profile},
    )
    if not response:
        raise RuntimeError("Non 2xx status code.")

    return [person(p["element"]) for p in response.json()["picks"]]


def my_team() -> structures.Squad:
    """
    Constructs and returns a Squad object representing the user's current team.
    """

    def fuzzyed_equal(player: structures.Player, persona: Persona) -> bool:
        personawebname = persona.webname.casefold()
        playerwebname = player.webname.casefold()

        personaname = persona.combined.casefold()
        playername = player.name.casefold()
        return (
            personawebname == playerwebname
            and all(name in personaname for name in playername.split())
            and all(name in playername for name in personaname.split())
        )

    return structures.Squad(
        [p for p in players() if any(fuzzyed_equal(p, pck) for pck in picks())]
    )



from __future__ import annotations

import argparse
import collections
import itertools
from typing import Generator, NamedTuple, Sequence

from lazyfpl import fetch, helpers, optimizer, structures


class Transfer(NamedTuple):
    bought: optimizer.PositionCombination
    sold: optimizer.PositionCombination


def display(trans: Transfer) -> None:
    """Displays the changes between the old and new player sequences,
    including transfers in and out."""
    sold = sorted(
        (p for p in trans.sold.players),
        key=lambda x: helpers.position_order(x.position),
    )
    bought = sorted(
        (p for p in trans.bought.players),
        key=lambda x: helpers.position_order(x.position),
    )

    max_len_in_name = max(len(p.webname) for p in bought)
    max_len_in_team = max(len(p.team) for p in bought)

    max_len_out_name = max(len(p.webname) for p in sold)
    max_len_out_team = max(len(p.team) for p in sold)

    print("-" * 75)

    for s, b in zip(sold, bought):
        print(
            f"{s.position}: {s.webname:<{max_len_out_name}} "
            f"- {s.team:<{max_len_out_team}} {s.xP or 0.0:<5.1f}"
            "  -->>  "
            f"{b.webname:<{max_len_in_name}} - "
            f"{b.team:<{max_len_in_team}} {b.xP or 0.0:.1f}"
        )
    print(f"xP gain: {(trans.bought.xP-trans.sold.xP):.1f}")


def transfer(
    current: Sequence[structures.Player],
    pool: Sequence[structures.Player],
    add: Sequence[structures.Player],
    remove: Sequence[structures.Player],
    max_transfers: int,
    max_budget: int = 1000,
    max_players_per_team: int = 3,
) -> Generator[Transfer, None, None]:
    """Generates transfer options for a given squad within specified
    constraints and preferences."""

    sold = {
        n: tuple(
            optimizer.PositionCombination(
                helpers.squad_price(c), helpers.squad_xP(c), c
            )
            for c in sorted(
                itertools.combinations(current, n),
                key=helpers.squad_price,
            )
        )
        for n in range(1, max_transfers + 1)
    }

    if remove:
        for n, combinations in sold.items():
            sold[n] = tuple(
                c for c in combinations if all(r in c.players for r in remove)
            )

    pool = [p for p in pool if p not in current]
    bought = {
        n: tuple(
            optimizer.PositionCombination(
                helpers.squad_price(c), helpers.squad_xP(c), c
            )
            for c in sorted(
                itertools.combinations(pool, n),
                key=helpers.squad_price,
            )
        )
        for n in range(1, max_transfers + 1)
    }
    if add:
        for n, combinations in bought.items():
            bought[n] = tuple(
                c for c in combinations if all(a in c.players for a in add)
            )

    min_bought_price = {
        n: min(c.price for c in combinations) for n, combinations in bought.items()
    }

    for n, combinations in sold.items():
        print(f"Sell combinations:   {n} - {len(combinations)}")

    for n, combinations in bought.items():
        print(f"Bought combinations: {n} - {len(combinations)}")

    current_squad_price = helpers.squad_price(current)
    current_team_tally = collections.Counter(p.team for p in current)

    for nin, bought_combinations in bought.items():
        for sold_combination in sold[nin]:
            if (
                current_squad_price - sold_combination.price + min_bought_price[nin]
                > max_budget
            ):
                continue

            for bought_combination in bought_combinations:
                if bought_combination.xP < sold_combination.xP:
                    continue

                if (
                    current_squad_price
                    - sold_combination.price
                    + bought_combination.price
                    > max_budget
                ):
                    continue

                team_tally = (
                    current_team_tally
                    + collections.Counter(p.team for p in bought_combination.players)
                    - collections.Counter(p.team for p in sold_combination.players)
                )
                if max(team_tally.values()) > max_players_per_team:
                    continue

                bought_sold_position_tally = collections.Counter(
                    p.position for p in bought_combination.players
                ) - collections.Counter(p.position for p in sold_combination.players)

                if bought_sold_position_tally:
                    continue

                yield Transfer(
                    bought_combination,
                    sold_combination,
                )


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Transfer picker.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--add",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--exclude",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-candidates",
        default=100,
        type=int,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-transfers",
        type=int,
        required=True,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        default=0.0,
        type=float,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-xp",
        default=0.0,
        type=float,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--remove",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    pool = [p for p in fetch.players() if p.xP is not None]

    if args.exclude:
        pool = [p for p in pool if p.webname not in args.exclude]
        pool = [p for p in pool if p.team not in args.exclude]

    if args.min_xp:
        pool = [p for p in pool if p.xP >= args.min_xp]

    if args.min_mtm:
        pool = [p for p in pool if p.mtm() >= args.min_mtm]

    if args.no_news:
        pool = [p for p in pool if not p.news]

    if args.add:
        pool += [p for p in fetch.players() if p.webname in args.add]

    print(">>> Pool")
    print(structures.Squad(pool))

    print("\n>>>> Current team")
    team = fetch.my_team()
    print(team)
    print()

    transfers = sorted(
        transfer(
            current=team.players,
            pool=list(set(pool)),
            add=[p for p in pool if p.webname in args.add] if args.add else [],
            remove=[p for p in team.players if p.webname in args.remove]
            if args.remove
            else [],
            max_transfers=args.max_transfers,
        ),
        key=lambda x: x.bought.xP - x.sold.xP,
    )[-100:]

    for trans in transfers:
        display(trans)


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
from itertools import chain, groupby

from lazyfpl import fetch, helpers

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--min-mtm",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-selected",
        "-mc",
        default=1_000,
        help=(
            "Player must be selected by at least this amunt of"
            "managers. (default: %(default)s)"
        ),
        type=int,
    )
    parser.add_argument(
        "--min-xp",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="Drop players with news attched to them. (default: %(default)s)",
    )
    parser.add_argument(
        "--top",
        "-t",
        default=None,
        help="Top N players per position. (default: %(default)s)",
        type=int,
    )
    args = parser.parse_args()

    pool = fetch.players()

    if args.no_news:
        pool = [p for p in pool if not p.news]

    pool = sorted(
        [
            p
            for p in pool
            if p.selected > args.min_selected
            and (p.xP or 0) > args.min_xp
            and p.mtm() > args.min_mtm
        ],
        key=lambda x: (
            -helpers.position_order(x.position),
            -(x.xP or 0) / (x.selected or 1),
        ),
    )

    pool = list(
        chain.from_iterable(
            list(x)[: args.top]
            for _, x in groupby(
                pool,
                key=lambda x: helpers.position_order(x.position),
            )
        )
    )

    print(
        helpers.tabulater(
            [
                p.display()
                for p in sorted(
                    pool,
                    key=lambda x: (
                        -helpers.position_order(x.position),
                        -(x.xP or 0),
                    ),
                )
            ],
        ),
    )



from __future__ import annotations

import argparse
import collections
import dataclasses
import functools
import itertools
import math
import pickle
import traceback
import typing

import more_itertools
import torch
from torch.utils.data import DataLoader as TorchDataLoader, Dataset as TorchDataset
from tqdm.std import tqdm

from lazyfpl import conf, database, fetch, populator, structures

if conf.debug:
    torch.set_printoptions(threshold=10_000)


@dataclasses.dataclass
class NormalizedFeatures:
    at_home: float
    minutes: float
    opponent_strength: float
    points: float
    team_strength: float
    opponent: tuple[float, ...]

    def flattend(self) -> tuple[float, ...]:
        def _flatter(obj: object) -> typing.Generator[float | int | bool, None, None]:
            if isinstance(obj, (float, int, bool)):
                yield obj
            elif isinstance(obj, (tuple, list, set)):
                for v in obj:
                    yield from _flatter(v)
            else:
                raise NotImplementedError(type(obj))

        return tuple(_flatter(dataclasses.astuple(self)))


@dataclasses.dataclass
class FeatureBundle:
    features: tuple[NormalizedFeatures, ...]
    target: float


class Net(torch.nn.Module):
    def __init__(
        self,
        nfeature: int,
        backtrace: int = conf.backtrace,
        scale_down: int = 8,
    ) -> None:
        super().__init__()
        self.nfeature = nfeature
        self.net = torch.nn.Sequential(
            torch.nn.Linear(nfeature * backtrace, nfeature // scale_down),
            torch.nn.BatchNorm1d(num_features=nfeature // scale_down),
            torch.nn.ELU(),
            torch.nn.Linear(nfeature // scale_down, 1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x.reshape(x.shape[0], -1)).squeeze()


@functools.cache
def onehot_team_name(team_name: str) -> tuple[float, ...]:
    teams = sorted({g.team for g in database.games()})
    enc = [0.0] * len(teams)
    enc[teams.index(team_name)] = 1.0
    assert math.isclose(sum(enc), 1.0)
    return tuple(enc)


def features(f: structures.Fixture) -> NormalizedFeatures:
    """Generates and returns normalized features for a given fixture."""
    assert f.points is not None
    points_scale = database.points()
    minutes_scale = database.minutes()
    return NormalizedFeatures(
        at_home=(f.at_home - 0.5) / 0.5,
        minutes=minutes_scale.unit_variance_normalization(f.minutes or 0),
        opponent_strength=(f.opponent_strength - 3) / 2,
        opponent=onehot_team_name(f.opponent),
        points=points_scale.unit_variance_normalization(f.points),
        team_strength=(f.team_strength - 3) / 2,
    )


class SequenceDataset(TorchDataset):
    def __init__(self, x: torch.Tensor, y: torch.Tensor) -> None:
        self.x, self.y = x, y

    def __len__(self) -> int:
        return self.x.shape[0]

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        return self.x[idx], self.y[idx]


def samples(
    fixtures: list[structures.Fixture],
    upsample: int,
    backtrace: int = conf.backtrace,
) -> typing.Iterator[FeatureBundle]:
    """Generates training samples from a list of fixtures, considering
    upsampling and backtrace length."""
    fixtures = [f for f in fixtures if not f.upcoming]
    fixtures = sorted(fixtures, key=lambda x: x.kickoff_time)
    # time --->
    min_ko = min(f.kickoff_time for f in fixtures)
    max_ko = max(f.kickoff_time for f in fixtures)

    if len(fixtures) < backtrace + 1:
        raise ValueError("To few samples.")

    for *context, target in more_itertools.sliding_window(fixtures, backtrace + 1):
        # Split window into context and target, context beeing the players previues
        # performences being used to predict the outcome of the upcoming match.
        assert len(context) == backtrace
        repat = max(
            (
                math.exp((target.kickoff_time - min_ko) / (max_ko - min_ko) * 2)
                / (math.e**2)
                * upsample
            ),
            1,
        )
        assert target.points is not None
        yield from itertools.repeat(
            FeatureBundle(
                features=tuple(features(c) for c in context),
                target=target.points,
            ),
            round(repat),
        )


def train(
    player: structures.Player,
    epochs: int,
    lr: float,
    upsample: int,
    batch_size: int,
) -> Net:
    """Trains a model for the given player using the specified parameters."""
    bundles = tuple(samples(player.fixutres, upsample, conf.backtrace))
    features = [[f.flattend() for f in b.features] for b in bundles]
    ds = SequenceDataset(
        x=torch.tensor(
            features,
            dtype=torch.float32,
        ),
        y=torch.tensor(
            tuple(b.target for b in bundles),
            dtype=torch.float32,
        ),
    )
    loader = TorchDataLoader(
        ds,
        batch_size=batch_size,
        shuffle=True,
    )

    loss_function = torch.nn.MSELoss()
    net = Net(ds[0][0].shape[-1])

    optimizer = torch.optim.SGD(net.parameters(), lr=lr)

    for _ in range(epochs):
        for x, y in loader:
            optimizer.zero_grad()
            output = net(x)
            assert output.shape == y.shape, (output.shape, y.shape)
            loss = loss_function(output, y)
            loss.backward()
            optimizer.step()
    return net


def load_model(player: structures.Player) -> Net:
    """Loads a trained model for the specified player."""
    pid = populator.player_id_fuzzer(player.name)
    if not pid:
        raise KeyError(player.name)
    if bts := database.load_model(pid):
        ms = pickle.loads(bts)
        n = Net(nfeature=ms["nfeature"])
        n.load_state_dict(ms["weights"])
        return n
    raise ValueError(f"No model for {player.name=} / {player.team=} / {pid=}.")


def save_model(player: structures.Player, m: Net) -> None:
    """Saves the trained model for the specified player."""
    pid = populator.player_id_fuzzer(player.name)
    if not pid:
        raise KeyError(player.name)
    database.save_model(
        pid,
        pickle.dumps(
            {
                "nfeature": m.nfeature,
                "weights": m.state_dict(),
            }
        ),
    )


def xP(
    player: structures.Player,
    lookahead: int = conf.lookahead,
    backtrace: int = conf.backtrace,
) -> float:
    """Calculates the expected points (xP) for a player based
    on their upcoming fixtures."""
    expected = list[float]()
    fixutres = sorted(player.fixutres, key=lambda x: x.kickoff_time)
    inference = [features(f).flattend() for f in fixutres if not f.upcoming][
        -backtrace:
    ]
    mtm = player.mtm()
    upcoming = [f for f in fixutres if f.upcoming]
    model = load_model(player).eval()
    with torch.no_grad():
        for nxt in upcoming[:lookahead]:
            points = (
                model(
                    torch.tensor(inference, dtype=torch.float32).unsqueeze(0),
                )
                .detach()
                .numpy()
            )
            expected.append(points)
            inference.pop(0)
            inference.append(
                features(
                    structures.Fixture(
                        **(
                            collections.ChainMap(
                                {"points": points, "minutes": mtm},
                                dataclasses.asdict(nxt),
                            )
                        )
                    )
                ).flattend()
            )

    if conf.debug:
        print(player.name, player.team, expected)

    return round(sum(expected), 1)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Player model trainer.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--epochs",
        type=int,
        default=5,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=0.01,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        type=int,
        default=0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--upsample",
        type=int,
        default=16,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=16,
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    players = [p for p in fetch.players() if p.mtm() >= args.min_mtm]
    max_webname = max(len(p.webname) for p in players)

    with tqdm(
        ascii=True,
        ncols=80,
        total=len(players),
        unit_scale=True,
    ) as bar:
        for player in players:
            bar.set_description_str(f"{player.webname:<{max_webname}}")
            try:
                m = train(
                    player,
                    epochs=args.epochs,
                    lr=args.lr,
                    upsample=args.upsample,
                    batch_size=args.batch_size,
                )
            except (
                IndexError,
                ValueError,
            ):
                ...
            except Exception as e:
                bar.write("".join(traceback.format_exception(e)))
            else:
                save_model(player, m)
                bar.write(
                    f"{xP(player):<6.1f} "
                    + f"{player.webname} "
                    + f"({player.team_short}) - "
                    + f"{player.str_upcoming_opponents()}"
                )
            finally:
                bar.update(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
import collections
import heapq
import itertools
from typing import Generator, NamedTuple

from tqdm.std import tqdm

from lazyfpl import constraints, fetch, helpers, structures


class PositionCombination(NamedTuple):
    price: int
    xP: float
    players: tuple[structures.Player, ...]


def position_combinations(
    pool: list[structures.Player],
    combinations: int,
) -> list[PositionCombination]:
    """Generates and returns sorted combinations of players with their
    total price and expected points."""
    assert combinations > 0
    return sorted(
        (
            PositionCombination(helpers.squad_price(c), helpers.squad_xP(c), c)
            for c in itertools.combinations(pool, combinations)
        ),
        key=lambda x: (x[1], -x[0]),
        reverse=True,
    )


def position_price_candidates(
    pool: list[structures.Player],
    topn: int,
) -> Generator[structures.Player, None, None]:
    """Iterates over players in a pool, returning the top
    players based on position and price."""
    for _, players in itertools.groupby(
        sorted(pool, key=lambda x: (x.position, x.price)),
        key=lambda x: (x.position, x.price),
    ):
        toadd = []
        team = set[str]()
        for player in sorted(players, key=lambda x: x.xP or 0, reverse=True):
            if player.team not in team:
                toadd.append(player)
                team.add(player.team)
        yield from toadd[:topn]


def must_include(
    combinations: list[PositionCombination],
    include: list[structures.Player],
) -> list[PositionCombination]:
    """Filters combinations to only include certain players."""
    return (
        [c for c in combinations if all(i in c[-1] for i in include)]
        if include
        else combinations
    )


def lineups_xp(
    gkp_combinations: list[PositionCombination],
    def_combinations: list[PositionCombination],
    mid_combinations: list[PositionCombination],
    fwd_combinations: list[PositionCombination],
    budget_lower: int = 900,
    budget_upper: int = 1_000,
    n_squads: int = 1_000,
    max_players_per_team: int = 3,
    score_decay: float = 0.995,
) -> list[structures.Squad]:
    """Generates the best possible lineups within given constraints."""
    # All combinations sorted from higest -> lowest xP.
    assert 0 < score_decay < 1

    total = (
        len(gkp_combinations)
        * len(def_combinations)
        * len(mid_combinations)
        * len(fwd_combinations)
    )

    print(f"Goalkeeper combinations: {len(gkp_combinations):.1e}")
    print(f"Defender   combinations: {len(def_combinations):.1e}")
    print(f"Midfielder combinations: {len(mid_combinations):.1e}")
    print(f"Forwarder  combinations: {len(fwd_combinations):.1e}")
    print(f"Total      combinations: {total:.1e}")

    if not total:
        return []

    best_squads = list[tuple[tuple[float, float, int], tuple[structures.Player, ...]]]()
    best_squad_xp = sum(
        (
            gkp_combinations[0][1],
            def_combinations[0][1],
            mid_combinations[0][1],
            fwd_combinations[0][1],
        )
    )

    max_mid_price = max(price for price, _, _ in mid_combinations)
    min_mid_price = min(price for price, _, _ in mid_combinations)
    max_mid_xp = max(xp for _, xp, _ in mid_combinations)

    max_def_price = max(price for price, _, _ in def_combinations)
    min_def_price = min(price for price, _, _ in def_combinations)
    max_def_xp = max(xp for _, xp, _ in def_combinations)

    sequence: int = 0

    with tqdm(
        ascii=True,
        leave=True,
        ncols=80,
        total=total,
        unit_scale=True,
    ) as bar:
        while len(best_squads) < min(n_squads, total) and best_squad_xp > 0:
            best_squad_xp *= score_decay
            bar.reset()
            bar.set_postfix_str(f"Squad xP cutoff: {best_squad_xp:.1f}")
            for gp, gxp, g in gkp_combinations:
                for fp, fxp, f in fwd_combinations:
                    bar.update(len(def_combinations) * len(mid_combinations))

                    if gxp + fxp + max_def_xp + max_mid_xp < best_squad_xp:
                        # Could use break, but messes up tqdm.
                        continue

                    if gp + fp + min_def_price + min_mid_price > budget_upper:
                        continue

                    if gp + fp + max_def_price + max_mid_price < budget_lower:
                        continue

                    if not constraints.team_constraint(g + f, max_players_per_team):
                        continue

                    for dp, dxp, d in def_combinations:
                        if gxp + fxp + dxp + max_mid_xp < best_squad_xp:
                            break

                        if gp + fp + dp + min_mid_price > budget_upper:
                            continue

                        if gp + fp + dp + max_mid_price < budget_lower:
                            continue

                        if not constraints.team_constraint(
                            g + f + d, max_players_per_team
                        ):
                            continue

                        for mp, mxp, m in mid_combinations:
                            if gxp + fxp + dxp + mxp < best_squad_xp:
                                break

                            if (
                                budget_lower
                                <= (price := mp + dp + fp + gp)
                                <= budget_upper
                                and constraints.team_constraint(
                                    squad := g + f + d + m,
                                    n=max_players_per_team,
                                )
                                and (oxp := helpers.overall_xP(squad)) > best_squad_xp
                                and not any(squad == s for _, s in best_squads)
                            ):
                                (
                                    heapq.heappushpop
                                    if len(best_squads) >= n_squads
                                    else heapq.heappush
                                )(
                                    best_squads,
                                    (
                                        (
                                            round(oxp, 1),
                                            price,
                                            sequence := sequence + 1,
                                        ),
                                        squad,
                                    ),
                                )

    return [
        structures.Squad(heapq.heappop(best_squads)[-1])
        for _ in range(len(best_squads))
    ]


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Lineup optimizer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--budget-lower",
        type=int,
        default=900,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--budget-upper",
        type=int,
        default=1_000,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--gkp-def-not-same-team",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--include",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--keep-squad",
        type=int,
        default=1_000,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-def-per-team",
        type=int,
        default=3,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-players-per-team",
        type=int,
        default=3,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-xp",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--remove",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--top-position-price",
        type=int,
        default=0,
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    pool = [p for p in fetch.players() if p.xP is not None]
    pool = [p for p in pool if p.mtm() >= args.min_mtm and p.xP >= args.min_xp]

    if args.no_news:
        pool = [p for p in pool if not p.news]

    player_names = {p.webname.lower() for p in pool}
    remove: set[str] = {r.lower() for r in args.remove if r.lower() in player_names}
    pool = [p for p in pool if p.webname.lower() not in remove]

    team_names = {p.team.lower() for p in pool}
    remove_team: set[str] = {r.lower() for r in args.remove if r.lower() in team_names}
    pool = [p for p in pool if p.team.lower() not in remove_team]

    if args.top_position_price:
        pool = list(
            set(
                position_price_candidates(
                    pool=pool,
                    topn=args.top_position_price,
                )
            )
        )

    include = list(
        {
            p
            for p in fetch.players()
            if p.webname in args.include
            or p.name in args.include
            or p.team in args.include
        }
    )
    pool += include

    # Just in case
    pool = list(set(pool))

    print(structures.Squad(pool))
    squads = lineups_xp(
        gkp_combinations=must_include(
            position_combinations(
                pool=[p for p in pool if p.position == "GKP"],
                combinations=2,
            ),
            [p for p in include if p.position == "GKP"],
        ),
        def_combinations=must_include(
            position_combinations(
                pool=[p for p in pool if p.position == "DEF"],
                combinations=5,
            ),
            [p for p in include if p.position == "DEF"],
        ),
        mid_combinations=must_include(
            position_combinations(
                pool=[p for p in pool if p.position == "MID"],
                combinations=5,
            ),
            [p for p in include if p.position == "MID"],
        ),
        fwd_combinations=must_include(
            position_combinations(
                pool=[p for p in pool if p.position == "FWD"],
                combinations=3,
            ),
            [p for p in include if p.position == "FWD"],
        ),
        budget_lower=args.budget_lower,
        budget_upper=args.budget_upper,
        max_players_per_team=args.max_players_per_team,
        n_squads=args.keep_squad,
    )

    if args.gkp_def_not_same_team:
        squads = [s for s in squads if constraints.gkp_def_same_team(s.players)]

    if args.max_def_per_team:
        squads = [
            s
            for s in squads
            if max(
                collections.Counter(p.team for p in s if p.position == "DEF").values()
            )
            <= args.max_def_per_team
        ]

    # Sort lineups by TSscore.
    squads = sorted(
        squads,
        key=lambda x: -helpers.tsscore(x.players),
    )

    print("\n\n".join(str(s) for s in squads))

    mincxp = min(s.CxP() for s in squads)
    maxxcp = max(s.CxP() for s in squads)
    minss = min(s.sscore() for s in squads)
    maxss = max(s.sscore() for s in squads)
    mints = min(s.tsscore() for s in squads)
    maxts = max(s.tsscore() for s in squads)

    print("")
    print(
        f"Min CxP: {mincxp:.1f}",
        f"Max CxP: {maxxcp:.1f}",
        f"Max-Min CxP: {(maxxcp-mincxp):.1f}",
    )
    print(
        f"Min ss: {minss:d}",
        f"Max ss: {maxss:d}",
        f"Max-Min ss: {(maxss-minss):d}",
    )
    print(
        f"Min ts: {mints:.2f}",
        f"Max ts: {maxts:.2f}",
        f"Max-Min ts: {(maxts-mints):.2f}",
    )


if __name__ == "__main__":
    main()
